// This is the panel test for use during construction.
// Specify the panel you wish to test below = 0 - 7; 0 is closest to the Mega

int mypanel = 0;  // set panel for test here.

// content of the cube is stored in these two arrays
// they are stored in two different arrays to facilitate the transfer of whole bytes directly into output ports during refresh
volatile byte anodes[4][8];
volatile byte cathodes[4][8];

// *********************************Setup ******************************
void setup() {
  // set 48 pins to output and initialize low
  for (int x = 2; x < 14; x++) {
    pinMode(x, OUTPUT);
    digitalWrite(x, LOW);
  }
  for (int x = 18; x < 54; x++) {
    pinMode(x, OUTPUT);
    digitalWrite(x, LOW);
  }

  // TIMER 1 for interrupt frequency 80 Hz = refresh every 12.5 milliseconds
  cli(); // stop interrupts
  TCCR1A = 0; // set entire TCCR1A register to 0
  TCCR1B = 0; // same for TCCR1B
  TCNT1  = 0; // initialize counter value to 0
  // set compare match register for 80 Hz increments
  OCR1A = 24999; // = 16000000 / (8 * 80) - 1 (must be <65536)
  // turn on CTC mode
  TCCR1B |= (1 << WGM12);
  // Set CS12, CS11 and CS10 bits for 8 prescaler
  TCCR1B |= (0 << CS12) | (1 << CS11) | (0 << CS10);
  // enable timer compare interrupt
  TIMSK1 |= (1 << OCIE1A);
  sei(); // allow interrupts
  delay(1);
}
// ***************************** End Setup ************************************

// *****************************  Main Loop ************************************
void loop() {

  for (int y = 0; y < 8; y++) {
    for (int z = 0; z < 8; z++) {
      ledOn(mypanel, y, z);
      delay(200);
      ledOff(mypanel, y, z);
    }
  }

}
// ***************************** End Main Loop *****************************************

// *******************************  Start of General Support Routines ************************

// this routine pushes the content of each layer down to the next
// then clears the top layer
void copyDown() {
  for (int z = 0; z < 7; z++) { // copy each layer to the one below it
    for (int x = 0; x < 4; x++) {
      for (int y = 0; y < 8; y++) {
        cathodes[x][z] = cathodes[x][z + 1];
        anodes[x][z] = anodes[x][z + 1];
      }
    }
  }
  for (int x = 0; x < 4; x++) {  // then clear the top layer
    cathodes[x][7] = 0;
    anodes[x][7] = 0;
  }
}

// turns everything off
void clearCube() {
  memset(anodes, 0, 32 * sizeof anodes[0][0]);
  memset(cathodes, 0, 32 * sizeof cathodes[0][0]);
}

// turn on an LED at x,y,z
void ledOn(byte x, byte y, byte z) {
  if (y % 2) { // y is odd, so cathode is coming down to main board
    if (x % 2) { // if x is odd, we are storing status in upper 4 bits
      cathodes[x >> 1][z] |= (1 << 4 + (y >> 1));
    }
    else {  // x is even, so we are storing status in lower 4 bits
      cathodes[x >> 1][z] |= (1 << (y >> 1));
    }
  }
  else {  // y is even, so anode is coming down to mainboard
    if (x % 2) { // if x is odd, we are storing status in upper 4 bits
      anodes[x >> 1][z] |= (1 << 4 + (y >> 1));
    }
    else {  //  x is even, we are storing status in lower 4 bits
      anodes[x >> 1][z] |= (1 << (y >> 1));
    }
  }
}

//turn off an LED at x,y,z
void ledOff(byte x, byte y, byte z) {
  if (y % 2) { // y is odd, so cathode is coming down to main board
    if (x % 2) { // if x is odd, we are storing status in upper 4 bits
      cathodes[x >> 1][z] &= ~(1 << 4 + (y >> 1));
    }
    else {  // x is even, so we are storing status in lower 4 bits
      cathodes[x >> 1][z] &= ~(1 << (y >> 1));
    }
  }
  else {  // y is even, so anode is coming down to mainboard
    if (x % 2) { // if x is odd, we are storing status in upper 4 bits
      anodes[x >> 1][z] &= ~(1 << 4 + (y >> 1));
    }
    else {  //  x is even, we are storing status in lower 4 bits
      anodes[x >> 1][z] &= ~(1 << (y >> 1));
    }
  }
}


// show cube for a multiples of 5 ms before clearing it.
void showCube(int mytime) {
  delay(10 * mytime);
  clearCube();
}

void pause() {  // pause between animations
  clearCube();
  delay(500);
}

// *******************************  End of General Support Routines ************************

// *******************************  Start of Refresh Timer Interrupt ************************
ISR(TIMER1_COMPA_vect) {
  for (int layer = 0; layer < 8; layer++) { // for each layer
    // setup anodes for this layer - high is on
    PORTA = anodes[0][layer];
    PORTB = anodes[1][layer];
    PORTC = anodes[2][layer];
    PORTL = anodes[3][layer];
    delayMicroseconds(3);  // settling time
    // pull the cathodes in this layer low for 500 µsec. by pulling layer pin high
    if      (layer == 0) PORTG |= (1 << PORTG0);
    else if (layer == 1) PORTG |= (1 << PORTG1);
    else if (layer == 2) PORTG |= (1 << PORTG2);
    else if (layer == 3) PORTE |= (1 << PORTE3);
    else if (layer == 4) PORTH |= (1 << PORTH3);
    else if (layer == 5) PORTH |= (1 << PORTH4);
    else if (layer == 6) PORTH |= (1 << PORTH5);
    else if (layer == 7) PORTH |= (1 << PORTH6);
    delayMicroseconds(500);
    PORTG = 0; // all cathode layer pins low
    PORTE = 0;
    PORTH = 0;
    delayMicroseconds(3);  // settling time
    // now setup cathodes for this layer - low is on
    PORTA = ~ cathodes[0][layer];
    PORTB = ~ cathodes[1][layer];
    PORTC = ~ cathodes[2][layer];
    PORTL = ~ cathodes[3][layer];
    delayMicroseconds(3);  // settling time
    // pull the anodes in this layer high for 500 µsec. by pulling layer pin high
    if (layer == 0)      PORTD |= (1 << PORTD0);
    else if (layer == 1) PORTD |= (1 << PORTD1);
    else if (layer == 2) PORTD |= (1 << PORTD2);
    else if (layer == 3) PORTD |= (1 << PORTD3);
    else if (layer == 4) PORTD |= (1 << PORTD7);
    else if (layer == 5) PORTG |= (1 << PORTG5);
    else if (layer == 6) PORTE |= (1 << PORTE5);
    else if (layer == 7) PORTE |= (1 << PORTE4);
    delayMicroseconds(500);
    PORTD = 0; // all anode layer pins low
    PORTG = 0;
    PORTE = 0;
    delayMicroseconds(3);  // settling time
  }
}
// *******************************  End of Refresh Timer Interrupt ************************
