// This tab/file contains General Support Subroutines and the Refresh Timer ISR


// *******************************  Start of General Support Routines ************************

// this routine pushes the content of each layer down to the next
// then clears the top layer
void copyDown() {
  for (int z = 0; z < 7; z++) { // copy each layer to the one below it
    for (int x = 0; x < 4; x++) {
      for (int y = 0; y < 8; y++) {
        cathodes[x][z] = cathodes[x][z + 1];
        anodes[x][z] = anodes[x][z + 1];
      }
    }
  }
  for (int x = 0; x < 4; x++) {  // then clear the top layer
    cathodes[x][7] = 0;
    anodes[x][7] = 0;
  }
}

// turns everything off by clearing the two primary arrays
void clearCube() {
  memset(anodes, 0, sizeof anodes);
  memset(cathodes, 0, sizeof cathodes);
}

// check if this is a valid location within the cube
bool inrange(int x, int y, int z) {
  if (x > -1 && x < 8 && y > -1 && y < 8 && z > -1 && z < 8) return true;
  else return false;
}

boolean ledStatus(byte x, byte y, byte z) {
  if (y % 2) { // y is odd, so cathode is coming down to main board
    if (x % 2) { // if x is odd, we are storing status in upper 4 bits
      if (((cathodes[x >> 1][z]) >> (4+(y >> 1)))&1)return true;
      else return false; 
    }
    else {  // x is even, so we are storing status in lower 4 bits
      if (((cathodes[x >> 1][z])>> (y >> 1))&1)return true;
      else return false;
    }
  }
  else {  // y is even, so anode is coming down to mainboard
    if (x % 2) { // if x is odd, we are storing status in upper 4 bits
      if (((anodes[x >> 1][z]) >> (4+(y >> 1)))&1)return true;
      else return false; 
    }
    else {  //  x is even, we are storing status in lower 4 bits
      if (((anodes[x >> 1][z])>> (y >> 1))&1)return true;
      else return false;
    }
  }
}

// turn on an LED at x,y,z
void ledOn(byte x, byte y, byte z) {
  if (y % 2) { // y is odd, so cathode is coming down to main board
    if (x % 2) { // if x is odd, we are storing status in upper 4 bits
      cathodes[x >> 1][z] |= (1 << 4 + (y >> 1));
    }
    else {  // x is even, so we are storing status in lower 4 bits
      cathodes[x >> 1][z] |= (1 << (y >> 1));
    }
  }
  else {  // y is even, so anode is coming down to mainboard
    if (x % 2) { // if x is odd, we are storing status in upper 4 bits
      anodes[x >> 1][z] |= (1 << 4 + (y >> 1));
    }
    else {  //  x is even, we are storing status in lower 4 bits
      anodes[x >> 1][z] |= (1 << (y >> 1));
    }
  }
}

//turn off an LED at x,y,z
void ledOff(byte x, byte y, byte z) {
  if (y % 2) { // y is odd, so cathode is coming down to main board
    if (x % 2) { // if x is odd, we are storing status in upper 4 bits
      cathodes[x >> 1][z] &= ~(1 << 4 + (y >> 1));
    }
    else {  // x is even, so we are storing status in lower 4 bits
      cathodes[x >> 1][z] &= ~(1 << (y >> 1));
    }
  }
  else {  // y is even, so anode is coming down to mainboard
    if (x % 2) { // if x is odd, we are storing status in upper 4 bits
      anodes[x >> 1][z] &= ~(1 << 4 + (y >> 1));
    }
    else {  //  x is even, we are storing status in lower 4 bits
      anodes[x >> 1][z] &= ~(1 << (y >> 1));
    }
  }
}


// show cube for a multiples of 5 ms before clearing it.
void showCube(int mytime) {
  delay(10 * mytime);
  clearCube();
}

// this is used to create separation between animations and clean everything up
void pause() {  // pause between animations
  clearCube();
  clearBufferCube();
  delay(500);
}

// The next six routines are for rotation of the entire content of the cube

// This routine rotates the content in the buffer_cube by a full revolution
// the specified number rotations in steps of specified stepAngle in degrees.
// The higher the stepAngle, the faster it will rotate.
void rotateCube(int times, int stepAngle) {
  rotation = stepAngle * .0174532; // convert rotate to radians
  int steps = (int)((6.28318 / abs(rotation)) * times) + 1;
  for (int x = 0; x < steps; x++) {
    myangle = myangle + rotation;
    if (rotation > 0 && myangle > 6.28318) {
      myangle = 0;
    }
    if (rotation < 0 && myangle <= 0) {
      myangle = 6.28318;
    }
    rotateAll(myangle);
  }
  clearCube();
}

// This routine rotates all layers by the specified angle and transfers it to the cube.
void rotateAll(float myangle) {
  for (byte layer = 0; layer < 8; layer++) {
    rotateLayer(myangle, layer);
  }
}

// This routine rotates one layer in the buffer_cube by the specified angle and transfers it
// to the cube.  It rotates the entire layer using polar coordinates.  Math is facilitated
// using the tables that are found after the font table in the Font_Table tab.
void rotateLayer(float myangle, int layer) {
  for (byte column = 0; column < 8; column++) { // clear this layer before proceeding
    for (byte panel = 0; panel < 8; panel++) {
      ledOff(column, panel, layer);
    }
  }
  for (byte column = 0; column < 8; column++) { // scan thru every column
    for (byte panel = 0; panel < 8; panel++) { // scan thru every panel
      float mydistance = distance[column][panel];
      float myNewAngle = angle[column][panel] + myangle;
      if (myNewAngle > 6.28318) {
        myNewAngle = myNewAngle - 6.28318;
      }
      int angleLookup = int(myNewAngle * 20 + .5);
      int newColumn = int(4 + (sin_cos[angleLookup][0] * mydistance));
      int newPanel = int(4 + (sin_cos[angleLookup][1] * mydistance));
      if (newColumn < 8 && newColumn > -1 && newPanel < 8 && newPanel > -1) {
        if (buffer_cube[column][panel][layer] > 0) ledOn(newColumn, newPanel, layer);
      }
    }
  }
}

void buffer_LEDon(int x, int y, int z) {
  buffer_cube[x][y][z] = 1;
}
void buffer_LEDoff(int x, int y, int z) {
  buffer_cube[x][y][z] = 0;
}
void clearBufferCube() {
  memset(buffer_cube, 0, sizeof buffer_cube);
}

// *******************************  End of General Support Routines ************************

// *******************************  Start of Refresh Timer Interrupt ************************
ISR(TIMER1_COMPA_vect) {
  for (int layer = 0; layer < 8; layer++) { // for each layer
    // setup anodes for this layer - high is on
    PORTA = anodes[0][layer];
    PORTB = anodes[1][layer];
    PORTC = anodes[2][layer];
    PORTL = anodes[3][layer];
    delayMicroseconds(3);  // settling time
    // pull the cathodes in this layer low for 500 µsec. by pulling layer pin high
    if      (layer == 0) PORTG |= (1 << PORTG0);
    else if (layer == 1) PORTG |= (1 << PORTG1);
    else if (layer == 2) PORTG |= (1 << PORTG2);
    else if (layer == 3) PORTE |= (1 << PORTE3);
    else if (layer == 4) PORTH |= (1 << PORTH3);
    else if (layer == 5) PORTH |= (1 << PORTH4);
    else if (layer == 6) PORTH |= (1 << PORTH5);
    else if (layer == 7) PORTH |= (1 << PORTH6);
    delayMicroseconds(500);
    PORTG = 0; // all cathode layer pins low
    PORTE = 0;
    PORTH = 0;
    delayMicroseconds(3);  // settling time
    // now setup cathodes for this layer - low is on
    PORTA = ~ cathodes[0][layer];
    PORTB = ~ cathodes[1][layer];
    PORTC = ~ cathodes[2][layer];
    PORTL = ~ cathodes[3][layer];
    delayMicroseconds(3);  // settling time
    // pull the anodes in this layer high for 500 µsec. by pulling layer pin high
    if (layer == 0)      PORTD |= (1 << PORTD0);
    else if (layer == 1) PORTD |= (1 << PORTD1);
    else if (layer == 2) PORTD |= (1 << PORTD2);
    else if (layer == 3) PORTD |= (1 << PORTD3);
    else if (layer == 4) PORTD |= (1 << PORTD7);
    else if (layer == 5) PORTG |= (1 << PORTG5);
    else if (layer == 6) PORTE |= (1 << PORTE5);
    else if (layer == 7) PORTE |= (1 << PORTE4);
    delayMicroseconds(500);
    PORTD = 0; // all anode layer pins low
    PORTG = 0;
    PORTE = 0;
    delayMicroseconds(3);  // settling time
  }
}
// *******************************  End of Refresh Timer Interrupt ************************
